///////////////////////////////////////////////////////////////////////////////
// File: ServerMain.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2004 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"
#include "JetByteTools/Admin/WinsockWrapper.h"

#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/DebugTrace.h"

#include "SocketServer.h"

#include "JetByteTools/SocketTools/FullAddress.h"

#include "ServerCommon/IOPool.h"
#include "ServerCommon/BufferAllocator.h"
#include "ServerCommon/CommandLine.h"
#include "ServerCommon/SimpleServerShutdownHandler.h"

#pragma hdrstop

#include "JetByteTools/CoreTools/StringConverter.h"
#include "JetByteTools/CoreTools/GlobalErrorHandler.h"

#include "JetByteTools/SocketTools/ServerCollection.h"
#include "JetByteTools/SocketTools/FullAddress.h"
#include "JetByteTools/SocketTools/ConnectionLimiter.h"

#include "JetByteTools/IOTools/AsyncFileLog.h"

#include "JetByteTools/SocketTools/AddressTypeIPv4.h"
#include "JetByteTools/SocketTools/AddressInfo.h"
#include "JetByteTools/SocketTools/AddressRenderer.h"
#include "JetByteTools/SocketTools/StreamSocketAllocator.h"

#include <iostream>

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::_tstring;
using JetByteTools::Core::CException;
using JetByteTools::Core::OutputEx;
using JetByteTools::Core::CDebugTrace;
using JetByteTools::Core::CStringConverter;
using JetByteTools::Core::CGlobalErrorHandler;

using JetByteTools::Socket::CServerCollection;
using JetByteTools::Socket::CFullAddress;
using JetByteTools::Socket::CAddressTypeIPv4;
using JetByteTools::Socket::CAddressType;
using JetByteTools::Socket::CConnectionLimiter;
using JetByteTools::Socket::ListenBacklog;
using JetByteTools::Socket::CStreamSocketAllocator;
using JetByteTools::Socket::CAddressInfo;
using JetByteTools::Socket::CAddressRenderer;

using JetByteTools::IO::CAsyncFileLog;

using std::cerr;
using std::endl;
using std::string;

///////////////////////////////////////////////////////////////////////////////
// Program entry point
///////////////////////////////////////////////////////////////////////////////

int main(int argc, const char *argv[ ])
{
   CGlobalErrorHandler errorHandler;

   int result = 0;

   try
   {
      CSimpleServerShutdownHandler shutdownHandler;

      CAsyncFileLog log(_T("MultiEchoServer.log"));

      CDebugTrace::LogInstaller logInstaller(log);

      try
      {
         CCommandLine commandLine(argc, argv, _T("MultiEchoServer"), CCommandLine::TCPServer);

         if (commandLine.Parse())
         {
            CIOPool pool(
               commandLine.NumberOfIOThreads(),
               commandLine.DisplayDebug());

            pool.Start();

            CStreamSocketAllocator socketAllocator(
               commandLine.SocketPoolSize(),
               commandLine.MaxBuffersPooledPerSocket(),
               commandLine.MaxConsecutiveOperationsOnOneSocket());

            CBufferAllocator bufferAllocator(
               commandLine.BufferSize(),
               commandLine.BufferPoolSize(),
               commandLine.DisplayDebug());

            const ListenBacklog listenBacklog = commandLine.ListenBacklog();

            CConnectionLimiter connectionLimiter(commandLine.MaxConnections());

            CServerCollection servers;

            unsigned short port = commandLine.Port();

            const CAddressType addressPreference = CAddressTypeIPv4::instance;

            servers.AddServer(new CSocketServer(
               commandLine.NoWelcomeMessage() ? "" : "Welcome to echo server 1\r\n",
               CFullAddress(commandLine.Server(), port, addressPreference),
               listenBacklog,
               pool,
               socketAllocator,
               bufferAllocator,
               connectionLimiter));

            port++;

            servers.AddServer(new CSocketServer(
               commandLine.NoWelcomeMessage() ? "" : "Welcome to echo server 2\r\n",
               CFullAddress(commandLine.Server(), port, addressPreference),
               listenBacklog,
               pool,
               socketAllocator,
               bufferAllocator,
               connectionLimiter));

            port++;

            servers.AddServer(new CSocketServer(
               commandLine.NoWelcomeMessage() ? "" : "Welcome to echo server 3\r\n",
               CFullAddress(commandLine.Server(), port, addressPreference),
               listenBacklog,
               pool,
               socketAllocator,
               bufferAllocator,
               connectionLimiter));

            port++;

            servers.AddServer(new CSocketServer(
               commandLine.NoWelcomeMessage() ? "" : "Welcome to echo server 4\r\n",
               CFullAddress(commandLine.Server(), port, addressPreference),
               listenBacklog,
               pool,
               socketAllocator,
               bufferAllocator,
               connectionLimiter));

            port++;

            servers.AddServer(new CSocketServer(
               commandLine.NoWelcomeMessage() ? "" : "Welcome to echo server 5\r\n",
               CFullAddress(commandLine.Server(), port, addressPreference),
               listenBacklog,
               pool,
               socketAllocator,
               bufferAllocator,
               connectionLimiter));

            port++;

            servers.AddServer(new CSocketServer(
               commandLine.NoWelcomeMessage() ? "" : "Welcome to echo server 6\r\n",
               CFullAddress(commandLine.Server(), port, addressPreference),
               listenBacklog,
               pool,
               socketAllocator,
               bufferAllocator,
               connectionLimiter));

            // Multi-homed example; listen on different interfaces with different servers...

            CAddressRenderer renderer(false);

            CAddressInfo addressInfo;

            for (CAddressInfo::Iterator it = addressInfo.Begin(); it != addressInfo.End(); ++it)
            {
               const _tstring ipAddress = renderer.AsString(it);

               const string message = commandLine.NoWelcomeMessage() ? "" : "Welcome to echo server 7 on: " + CStringConverter::TtoA(ipAddress) + "\r\n";

               servers.AddServer(new CSocketServer(
                  message,
                  CFullAddress(ipAddress, port, addressPreference),
                  listenBacklog,
                  pool,
                  socketAllocator,
                  bufferAllocator));
            }

            servers.Start();

            servers.StartAcceptingConnections();

            shutdownHandler.SetServerControl(servers, commandLine.EventNamePrefix());

            shutdownHandler.WaitForShutdownRequest();

            servers.WaitForShutdownToComplete();

            pool.WaitForShutdownToComplete();

            socketAllocator.ReleaseSockets();

            bufferAllocator.Flush();
         }
         else
         {
            result = 1;
         }
      }
      catch(const CException &e)
      {
         OutputEx(_T("Exception: ") + e.GetDetails());

         result = 1;
      }
      catch(...)
      {
         OutputEx(_T("Unexpected exception"));

         result = 1;
      }
   }
   catch(const CException &e)
   {
      const _tstring message = _T("Exception: ") + e.GetDetails();

      cerr << CStringConverter::TtoA(message) << endl;

      result = 1;
   }
   catch(...)
   {
      cerr << "Unexpected exception" << endl;

      result = 1;
   }

   return result;
}

///////////////////////////////////////////////////////////////////////////////
// End of file: ServerMain.cpp
///////////////////////////////////////////////////////////////////////////////
