///////////////////////////////////////////////////////////////////////////////
// File: ServerMain.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2004 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"
#include "JetByteTools/Admin/WinsockWrapper.h"

#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/Win32Tools/GlobalName.h"
#include "JetByteTools/Win32Tools/ManualResetEvent.h"

#include "SocketServer.h"

#include "ServerCommon/IOPool.h"
#include "ServerCommon/BufferAllocator.h"
#include "ServerCommon/CommandLine.h"

#include "JetByteTools/SocketTools/FullAddress.h"
#include "JetByteTools/SocketTools/ConnectionLimiter.h"

#pragma hdrstop

#include "JetByteTools/Win32Tools/SecurityDescriptorAllowAll.h"
#include "JetByteTools/Win32Tools/SecurityAttributes.h"
#include "JetByteTools/CoreTools/StringConverter.h"
#include "JetByteTools/CoreTools/GlobalErrorHandler.h"

#include "JetByteTools/IOTools/AsyncFileLog.h"

#include "JetByteTools/SocketTools/AddressTypeIPv4.h"
#include "JetByteTools/SocketTools/StreamSocketAllocator.h"
#include "JetByteTools/SocketTools/SequencedStreamSocketAllocator.h"

#include <iostream>

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::_tstring;
using JetByteTools::Core::CException;
using JetByteTools::Core::OutputEx;
using JetByteTools::Win32::CGlobalName;
using JetByteTools::Win32::CManualResetEvent;
using JetByteTools::Core::CDebugTrace;
using JetByteTools::Win32::CSecurityDescriptorAllowAll;
using JetByteTools::Win32::CSecurityAttributes;
using JetByteTools::Core::CStringConverter;
using JetByteTools::Core::CGlobalErrorHandler;

using JetByteTools::Socket::CStreamSocketAllocator;
using JetByteTools::Socket::CSequencedStreamSocketAllocator;
using JetByteTools::Socket::CFullAddress;
using JetByteTools::Socket::CAddressTypeIPv4;
using JetByteTools::Socket::CAddressType;
using JetByteTools::Socket::CConnectionLimiter;
using JetByteTools::Socket::ListenBacklog;

using JetByteTools::IO::CAsyncFileLog;

using std::cerr;
using std::endl;

///////////////////////////////////////////////////////////////////////////////
// Program entry point
///////////////////////////////////////////////////////////////////////////////

int main(int argc, const char *argv[ ])
{
   CGlobalErrorHandler errorHandler;

   int result = 0;

   try
   {
      CAsyncFileLog log(_T("PacketEchoServer.log"));

      CDebugTrace::LogInstaller logInstaller(log);

      try
      {
         CCommandLine commandLine(argc, argv, _T("PacketEchoServer"), CCommandLine::TCPServer);

         if (commandLine.Parse())
         {
            CIOPool pool(
               commandLine.NumberOfIOThreads(),
               commandLine.DisplayDebug());

            pool.Start();

            CSequencedStreamSocketAllocator socketAllocator1(
               commandLine.SocketPoolSize(),
               commandLine.MaxBuffersPooledPerSocket(),
               commandLine.MaxConsecutiveOperationsOnOneSocket());

            CBufferAllocator bufferAllocator(
               commandLine.BufferSize(),
               commandLine.BufferPoolSize(),
               commandLine.DisplayDebug());

            const CAddressType addressPreference = CAddressTypeIPv4::instance;

            const CFullAddress address(
               commandLine.Server(),
               commandLine.Port(),
               addressPreference);

            const ListenBacklog listenBacklog = commandLine.ListenBacklog();

            CConnectionLimiter connectionLimiter(commandLine.MaxConnections());

            CSocketServer server1(
               commandLine.NoWelcomeMessage() ? "" : "Welcome to packet echo server (using sequence numbers)\r\n",
               address,
               listenBacklog,
               pool,
               socketAllocator1,
               bufferAllocator,
               connectionLimiter);

            server1.Start();
            server1.StartAcceptingConnections();

            CStreamSocketAllocator socketAllocator2(
               commandLine.SocketPoolSize(),
               commandLine.MaxBuffersPooledPerSocket(),
               commandLine.MaxConsecutiveOperationsOnOneSocket());

            const CFullAddress address2(
               commandLine.Server(),
               static_cast<unsigned short>(commandLine.Port() + 1),
               addressPreference);

            CSocketServer server2(
               commandLine.NoWelcomeMessage() ? "" : "Welcome to packet echo server (no sequence numbers)\r\n",
               address2,
               listenBacklog,
               pool,
               socketAllocator2,
               bufferAllocator,
               connectionLimiter);

            server2.Start();
            server2.StartAcceptingConnections();

            const _tstring eventNamePrefix = commandLine.EventNamePrefix();

            CSecurityDescriptorAllowAll sd;

            CSecurityAttributes sa(sd);

            CManualResetEvent serverStartedEvent(CGlobalName(eventNamePrefix + _T("ServerStarted")), &sa);
            CManualResetEvent serverStoppedEvent(CGlobalName(eventNamePrefix + _T("ServerStopped")), &sa);

            serverStartedEvent.Set();

            CManualResetEvent shutdownEvent(CGlobalName(eventNamePrefix + _T("ServerShutdown")), &sa);
            CManualResetEvent pauseResumeEvent(CGlobalName(eventNamePrefix + _T("ServerPauseResume")), &sa);

            HANDLE handlesToWaitFor[2];

            handlesToWaitFor[0] = shutdownEvent.GetWaitHandle();
            handlesToWaitFor[1] = pauseResumeEvent.GetWaitHandle();

            bool accepting = true;
            bool done = false;

            while (!done)
            {
               DWORD waitResult = ::WaitForMultipleObjects(2, handlesToWaitFor, false, INFINITE);

               if (waitResult == WAIT_OBJECT_0)
               {
                  done = true;
               }
               else if (waitResult == WAIT_OBJECT_0 + 1)
               {
                  if (accepting)
                  {
                     server1.StopAcceptingConnections();
                     server2.StopAcceptingConnections();
                  }
                  else
                  {
                     server1.StartAcceptingConnections();
                     server2.StartAcceptingConnections();
                  }

                  accepting = !accepting;
               }
               else
               {
                  OutputEx(_T("Unexpected result from WaitForMultipleObjects - exiting"));
                  done = true;
                  result = 1;
               }
            }

            server1.WaitForShutdownToComplete();
            server2.WaitForShutdownToComplete();

            pool.WaitForShutdownToComplete();

            socketAllocator1.ReleaseSockets();

            socketAllocator2.ReleaseSockets();

            bufferAllocator.Flush();

            serverStoppedEvent.Set();
         }
         else
         {
            result = 1;
         }
      }
      catch(const CException &e)
      {
         OutputEx(_T("Exception: ") + e.GetDetails());

         result = 1;
      }
      catch(...)
      {
         OutputEx(_T("Unexpected exception"));

         result = 1;
      }
   }
   catch(const CException &e)
   {
      const _tstring message = _T("Exception: ") + e.GetDetails();

      cerr << CStringConverter::TtoA(message) << endl;

      result = 1;
   }
   catch(...)
   {
      cerr << "Unexpected exception" << endl;

      result = 1;
   }

   return result;
}

///////////////////////////////////////////////////////////////////////////////
// End of file: ServerMain.cpp
///////////////////////////////////////////////////////////////////////////////
