///////////////////////////////////////////////////////////////////////////////
// File: SocketServer.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2004 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "SocketServer.h"

#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/Exception.h"

#include "JetByteTools/SocketTools/IStreamSocket.h"

#pragma hdrstop

#include "JetByteTools/IOTools/SmartBuffer.h"

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::Output;
using JetByteTools::Core::ToString;
using JetByteTools::Core::CException;
using JetByteTools::Core::DumpData;
using JetByteTools::Core::GetStringLength;

using JetByteTools::IO::IAllocateBuffers;
using JetByteTools::IO::IBuffer;
using JetByteTools::IO::CSmartBuffer;
using JetByteTools::IO::IIOPool;
using JetByteTools::IO::CSmartBuffer;

using JetByteTools::Socket::CStreamSocketServerEx;
using JetByteTools::Socket::IStreamSocket;
using JetByteTools::Socket::IAllocateStreamSockets;
using JetByteTools::Socket::IFullAddress;
using JetByteTools::Socket::IAddress;
using JetByteTools::Socket::ILimitConnections;
using JetByteTools::Socket::ListenBacklog;
using JetByteTools::Socket::NoZeroByteRead;

using std::string;

///////////////////////////////////////////////////////////////////////////////
// CSocketServer
///////////////////////////////////////////////////////////////////////////////

CSocketServer::CSocketServer(
   const string &welcomeMessage,
   const IFullAddress &address,
   const ListenBacklog listenBacklog,
   IIOPool &pool,
   IAllocateStreamSockets &socketAllocator,
   IAllocateBuffers &bufferAllocator,
   ILimitConnections &connectionLimiter)
   :  CStreamSocketServerEx(address, listenBacklog, *this, pool, socketAllocator, bufferAllocator, NoZeroByteRead, connectionLimiter),
      m_welcomeMessage(welcomeMessage)
{

}

CSocketServer::~CSocketServer()
{
   WaitForShutdownToComplete();
}

void CSocketServer::OnConnectionEstablished(
   IStreamSocket &socket,
   const IAddress & /*address*/)
{
   Output(_T("OnConnectionEstablished"));

   if (socket.TryWrite(m_welcomeMessage.c_str(), GetStringLength<DWORD>(m_welcomeMessage)))
   {
      socket.TryRead();
   }
}

void CSocketServer::OnReadCompleted(
   IStreamSocket &socket,
   CSmartBuffer &buffer)
{
   ProcessDataStream(socket, buffer);
}

void CSocketServer::EchoMessage(
   IStreamSocket &socket,
   CSmartBuffer &buffer) const
{
   //DEBUG_ONLY(Output(_T("Data Echoed -\r\n") + DumpData(buffer->GetMemory(), buffer->GetUsed(), 60, true)));

   socket.Write(buffer);
}

IBuffer::BufferSize CSocketServer::GetMinimumMessageSize() const
{
   return 1;
}

IBuffer::BufferSize CSocketServer::GetMessageSize(const IBuffer &buffer) const
{
   IBuffer::BufferSize messageSize = *buffer.GetMemory();

   return messageSize;
}

void CSocketServer::ProcessDataStream(
   IStreamSocket &socket,
   CSmartBuffer &buffer) const
{
   bool done;

   do
   {
      done = true;

      const IBuffer::BufferSize used = buffer->GetUsed();

      if (used < GetMinimumMessageSize())
      {
         // We have insufficient data in the buffer for a message

         socket.Read(buffer);
      }
      else
      {
         const IBuffer::BufferSize messageSize = GetMessageSize(buffer.GetRef());

         if (messageSize > buffer->GetSize())
         {
            throw CException(_T("CSocketServer::ProcessDataStream()"),
               _T("Error: Buffer too small\nExpecting: ") + ToString(messageSize) +
               _T("Got: ") + ToString(buffer->GetUsed()) + _T("\nBuffer size = ") +
               ToString(buffer->GetSize()) + _T("\nData = \n") +
               DumpData(buffer->GetMemory(), buffer->GetUsed(), 40));
         }
         else if (used < messageSize)
         {
            // We have insufficient data in the buffer for the message

            socket.Read(buffer);
         }
         else if (used == messageSize)
         {
            // we have a whole, distinct, message

            Output(_T("Got complete, distinct, message"));

            EchoMessage(socket, buffer);

            socket.Read();
         }
         else if (used > messageSize)
         {
            // we have a message, plus some more data

            Output(_T("Got message plus extra data"));

            // allocate a new buffer, copy the extra data into it and try again...

            CSmartBuffer extraData(buffer->SplitBufferFromBack(used - messageSize));

            EchoMessage(socket, buffer);

            buffer = extraData.Detach();

            // loop again, we may have another complete message in there...

            done = false;
         }
      }
   }
   while (!done);
}

///////////////////////////////////////////////////////////////////////////////
// End of file: SocketServer.cpp
///////////////////////////////////////////////////////////////////////////////
