///////////////////////////////////////////////////////////////////////////////
// File: IOPoolEx.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2012 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "IOPoolEx.h"

#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/ThreadLocalStorage.h"

#include "JetByteTools/Win32Tools/CriticalSection.h"

#pragma hdrstop

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::Output;
using JetByteTools::Core::OutputEx;
using JetByteTools::Core::_tstring;
using JetByteTools::Core::ToString;
using JetByteTools::Core::CThreadLocalStorage;
using JetByteTools::Win32::CCriticalSection;
using JetByteTools::Win32::ICriticalSection;

///////////////////////////////////////////////////////////////////////////////
// Static helper methods
///////////////////////////////////////////////////////////////////////////////

static DWORD GetTlsIndex();

static void ReleaseTlsIndex();

///////////////////////////////////////////////////////////////////////////////
// Static variables
///////////////////////////////////////////////////////////////////////////////

static CCriticalSection s_criticalSection;

static CThreadLocalStorage *s_pTls = nullptr;

static size_t s_ref = 0;

struct Counts
{
   Counts() : m_min(0xFFFFFFFF), m_max(0), m_total(0), m_calls(0)
   {
   }

   void Add(
      const DWORD numItems)
   {
      m_total += numItems;
      m_calls++;

      if (numItems > m_max)
      {
         m_max = numItems;
      }

      if (numItems < m_min)
      {
         m_min = numItems;
      }
   }

   DWORD m_min;
   DWORD m_max;

   size_t m_total;

   size_t m_calls;
};

///////////////////////////////////////////////////////////////////////////////
// CIOPoolEx
///////////////////////////////////////////////////////////////////////////////

CIOPoolEx::CIOPoolEx(
   const ULONG numEntriesPerCall,
   const ThreadCount numThreads,
   const bool log)
   :  JetByteTools::IO::CIOPoolEx(numEntriesPerCall, *this, numThreads),
      m_log(log),
      m_tlsIndex(GetTlsIndex())
{

}

CIOPoolEx::~CIOPoolEx()
{
   WaitForShutdownToComplete();

   ReleaseTlsIndex();
}

void CIOPoolEx::OnIOPoolError(
   const _tstring &message)
{
   OutputEx(_T("OnError - ") + message);
}

void CIOPoolEx::OnIOPoolThreadCreated()
{
   if (m_log)
   {
      Output(_T("OnIOPoolThreadCreated"));
   }
}

void CIOPoolEx::OnIOPoolThreadStarted()
{
   if (m_log)
   {
      Output(_T("OnIOPoolThreadStarted"));
   }

   CThreadLocalStorage tls(m_tlsIndex);

   tls.SetValue(reinterpret_cast<void*>(new Counts()));
}

void CIOPoolEx::OnIOPoolDispatch()
{
   if (m_log)
   {
      Output(_T("OnIOPoolDispatch"));
   }
}

void CIOPoolEx::OnIOPoolWorkItemsDequeued(
   const DWORD numWorkItems)
{
   if (m_log)
   {
      Output(_T("OnIOPoolWorkItemsDequeued: ") + ToString(numWorkItems));
   }

   CThreadLocalStorage tls(m_tlsIndex);

   Counts *pCounts = reinterpret_cast<Counts*>(tls.GetValue());

   pCounts->Add(numWorkItems);
}

void CIOPoolEx::OnIOPoolThreadBeginProcessing()
{
   if (m_log)
   {
      Output(_T("OnIOPoolThreadBeginProcessing"));
   }
}

void CIOPoolEx::OnIOPoolThreadEndProcessing()
{
   if (m_log)
   {
      Output(_T("OnIOPoolThreadEndProcessing"));
   }
}

void CIOPoolEx::OnIOPoolThreadStopped()
{
   if (m_log)
   {
      Output(_T("OnIOPoolThreadStopped"));
   }

   CThreadLocalStorage tls(m_tlsIndex);

   Counts *pCounts = reinterpret_cast<Counts*>(tls.GetValue());

   if (pCounts->m_calls)
   {
      OutputEx(
         _T("Thread: ") + ToString(GetCurrentThreadId()) +
         _T(" total: ") + ToString(pCounts->m_total) +
         _T(" average: ") + ToString(pCounts->m_total / pCounts->m_calls) +
         _T(" max: ") + ToString(pCounts->m_max) +
         _T(" min: ") + ToString(pCounts->m_min));
   }

   delete pCounts;
}

void CIOPoolEx::OnIOPoolThreadDestroyed()
{
   if (m_log)
   {
      Output(_T("OnIOPoolThreadDestroyed"));
   }
}

///////////////////////////////////////////////////////////////////////////////
// Static helper methods
///////////////////////////////////////////////////////////////////////////////

static DWORD GetTlsIndex()
{
   ICriticalSection::Owner lock(s_criticalSection);

   if (!s_pTls)
   {
      s_pTls = new CThreadLocalStorage();

      s_ref = 1;
   }
   else
   {
      s_ref++;
   }

   return s_pTls->GetIndex();
}

static void ReleaseTlsIndex()
{
   ICriticalSection::Owner lock(s_criticalSection);

   s_ref--;

   if (s_ref == 0)
   {
      delete s_pTls;
      s_pTls = nullptr;
   }
}

///////////////////////////////////////////////////////////////////////////////
// End of file: IOPoolEx.cpp
///////////////////////////////////////////////////////////////////////////////
