///////////////////////////////////////////////////////////////////////////////
// File: MessageAccumulator.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2016 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "MessageAccumulator.h"

#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/Exception.h"

#pragma hdrstop

#include "JetByteTools/IOTools/BufferBasedBufferAllocator.h"

#include <algorithm>    // for min

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::ToString;
using JetByteTools::Core::_tstring;
using JetByteTools::Core::CException;

using JetByteTools::IO::IBuffer;
using JetByteTools::IO::CSmartBuffer;
using JetByteTools::IO::CBufferBasedBufferAllocator;


///////////////////////////////////////////////////////////////////////////////
// CMessageAccumulator
///////////////////////////////////////////////////////////////////////////////

static const DWORD s_minMessageSize = sizeof(DWORD);  // The smallest possible
                                                      // packet is 4 bytes for
                                                      // the message length with
                                                      // a 0 length message.

CMessageAccumulator::CMessageAccumulator(
   const IBuffer::BufferSize maxMessageLength)
   :  m_maxMessageLength(maxMessageLength),
      m_messageLength(0),
      m_nextMessageSeq(0)
{
}

CMessageAccumulator::~CMessageAccumulator()
{
}

IBuffer::BufferSize CMessageAccumulator::GetMessageLength(
   const BYTE *pData) const
{
   const DWORD netOrder = *reinterpret_cast<const DWORD *>(pData);

   const IBuffer::BufferSize messageLength = sizeof(DWORD) + ntohl(netOrder);

   if (messageLength < s_minMessageSize)
   {
      throw CException(
         _T("CMessageAccumulator::GetMessageLength()"),
         _T("Message is too small: ") + ToString(messageLength) +
         _T(" min is: ") + ToString(s_minMessageSize));
   }

   if (messageLength > m_maxMessageLength)
   {
      throw CException(
         _T("CMessageAccumulator::GetMessageLength()"),
         _T("Message is too big: ") + ToString(messageLength) +
         _T(" max is: ") + ToString(m_maxMessageLength));
   }

   return messageLength;
}

CSmartBuffer CMessageAccumulator::Add(
   CSmartBuffer &buffer)
{
   const IBuffer::BufferSize numBytes = buffer->GetUsed();

   const BYTE *pData = buffer->GetMemory();

   if (m_messageLength == 0 && numBytes > s_minMessageSize)
   {
      m_messageLength = GetMessageLength(pData);
   }

   const IBuffer::BufferSize existingBytes = static_cast<IBuffer::BufferSize>(m_buffers.TotalBytes());

   if (existingBytes + numBytes > m_messageLength)
   {
      // too much in this buffer, just take the bytes we need...

      const IBuffer::BufferSize bytesUsed = m_messageLength - existingBytes;

      CBufferBasedBufferAllocator allocator(buffer.GetRef());

      // Add the remaining message bytes to the buffer chain, put them in
      // the last buffer in the chain, if possible, else allocate a new buffer

      m_buffers.AddData(
         allocator,
         pData,
         bytesUsed);

      // Mark the bytes that we removed as removed in the original buffer

      buffer->Consume(bytesUsed);
   }
   else
   {
      // add this buffer to the chain

      m_buffers.Add(buffer);
   }

   CSmartBuffer message;

   if (m_buffers.TotalBytes() == m_messageLength)
   {
      message = m_buffers.Detach();

      message->SetSequenceNumber(m_nextMessageSeq++);
   }

   return message;
}

///////////////////////////////////////////////////////////////////////////////
// End of file: MessageAccumulator.cpp
///////////////////////////////////////////////////////////////////////////////
