///////////////////////////////////////////////////////////////////////////////
// File: SocketServer.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2004 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "SocketServer.h"
#include "ThreadPool.h"
#include "MessageAccumulator.h"

#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/Exception.h"

#include "JetByteTools/SocketTools/IStreamSocket.h"
#include "JetByteTools/SocketTools/IAllocateSequencedStreamSockets.h"

#pragma hdrstop

#include <algorithm>    // for min

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::Output;
using JetByteTools::Core::ToString;
using JetByteTools::Core::_tstring;
using JetByteTools::Core::CException;
using JetByteTools::Core::IIndexedOpaqueUserData;
using JetByteTools::Core::GetStringLength;

using JetByteTools::IO::IIOPool;
using JetByteTools::IO::IBuffer;
using JetByteTools::IO::CSmartBuffer;
using JetByteTools::IO::IAllocateBuffers;

using JetByteTools::Socket::CFilteringStreamSocketServerEx;
using JetByteTools::Socket::IStreamSocket;
using JetByteTools::Socket::IAllocateSequencedStreamSockets;
using JetByteTools::Socket::IFullAddress;
using JetByteTools::Socket::ILimitConnections;
using JetByteTools::Socket::IAddress;
using JetByteTools::Socket::ListenBacklog;
using JetByteTools::Socket::SocketBufferSize;
using JetByteTools::Socket::ZeroByteReadConfiguration;

using std::min;

///////////////////////////////////////////////////////////////////////////////
// CSocketServer
///////////////////////////////////////////////////////////////////////////////

CSocketServer::CSocketServer(
   CThreadPool &businessLogicPool,
   const IBuffer::BufferSize maxMessageSize,
   const bool issueEarlyRead,
   const IFullAddress &address,
   const ListenBacklog listenBacklog,
   IIOPool &ioPool,
   IAllocateSequencedStreamSockets &socketAllocator,
   IAllocateBuffers &bufferAllocator,
   const SocketBufferSize recvBufferSize,
   const SocketBufferSize sendBufferSize,
   const ZeroByteReadConfiguration zeroByteReadConfiguration,
   ILimitConnections &connectionLimiter)
   :  CFilteringStreamSocketServerEx(
         address,
         listenBacklog,
         *this,
         ioPool,
         socketAllocator,
         bufferAllocator,
         recvBufferSize,
         sendBufferSize,
         zeroByteReadConfiguration,
         connectionLimiter),
      m_pool(businessLogicPool),
      m_maxMessageLength(maxMessageSize),
      m_issueEarlyRead(issueEarlyRead),
      m_userDataIndex(socketAllocator.RequestUserDataSlot(_T("CSocketServer - m_userDataIndex"))),
      m_readSequenceFilter(*this)
{

}

CSocketServer::~CSocketServer()
{
   CFilteringStreamSocketServerEx::WaitForShutdownToComplete();
}

void CSocketServer::OnConnectionEstablished(
   IStreamSocket &socket,
   const IAddress &address)
{
   Output(_T("OnConnectionEstablished"));

//#pragma warning(suppress: 28197) // Possibly leaking memory. No, we're not.
   CMessageAccumulator *pPerConnectionData = new CMessageAccumulator(m_maxMessageLength);

   SetPerConnectionData(socket, pPerConnectionData);

   m_pool.DispatchConnectionEstablished(socket, address);
}

void CSocketServer::OnSocketReleased(
   IIndexedOpaqueUserData &userData)
{
   Output(_T("OnSocketReleased"));

   CMessageAccumulator &perConnectionData = GetPerConnectionData(userData);

   SetPerConnectionData(userData, 0);

   delete &perConnectionData;

   m_pool.OnSocketReleased(userData);
}

void CSocketServer::OnReadCompleted(
   IStreamSocket &socket,
   CSmartBuffer &buffer)
{
   if (m_issueEarlyRead)
   {
      socket.TryRead();
   }

   //DEBUG_ONLY(Output(_T("Processing buffer: ") + ToString(buffer.Get()) + _T(": Seq:") + ToString(buffer->GetSequenceNumber())));

   ProcessDataStream(socket, buffer);

   if (!m_issueEarlyRead)
   {
      socket.TryRead();
   }
}

void CSocketServer::ProcessDataStream(
   IStreamSocket &socket,
   CSmartBuffer &buffer) const
{
   CMessageAccumulator &perConnectionData = GetPerConnectionData(socket);

   do
   {
      // Accumulate our message until we have a complete one and then process it...

      CSmartBuffer message(perConnectionData.Add(buffer));

      if (message.Get())
      {
         // message complete...

         ProcessCommand(socket, message);
      }
   }
   while (buffer.Get());
}

void CSocketServer::ProcessCommand(
   IStreamSocket &socket,
   CSmartBuffer &buffer) const
{
   m_pool.DispatchCompleteMessage(socket, buffer);
}

void CSocketServer::SetPerConnectionData(
   IIndexedOpaqueUserData &userData,
   CMessageAccumulator *pData) const
{
   userData.SetUserPointer(m_userDataIndex, pData);
}

CMessageAccumulator &CSocketServer::GetPerConnectionData(
   const IIndexedOpaqueUserData &userData) const
{
   return *reinterpret_cast<CMessageAccumulator*>(userData.GetUserPointer(m_userDataIndex));
}

///////////////////////////////////////////////////////////////////////////////
// End of file: SocketServer.cpp
///////////////////////////////////////////////////////////////////////////////
