///////////////////////////////////////////////////////////////////////////////
// File: ThreadPool.cpp
/////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2002 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"
#include "JetByteTools/Admin/WinsockWrapper.h"

#include "ThreadPool.h"

#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/IProvideUserData.h"
#include "JetByteTools/CoreTools/IIndexedOpaqueUserData.h"
#include "JetByteTools/CoreTools/StringConverter.h"

#include "JetByteTools/IOTools/IBuffer.h"
#include "JetByteTools/IOTools/BufferChain.h"
#include "JetByteTools/IOTools/CustomPoolSmartBuffer.h"
#include "JetByteTools/IOTools/IAllocateBuffers.h"

#include "JetByteTools/SocketTools/IStreamSocket.h"
#include "JetByteTools/SocketTools/AddressRenderer.h"
#include "JetByteTools/SocketTools/SmartStreamSocket.h"
#include "JetByteTools/SocketTools/IAllocateSequencedStreamSockets.h"

#pragma hdrstop

#include <algorithm>    // for min

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Milliseconds;

using JetByteTools::Core::Output;
using JetByteTools::Core::OutputEx;
using JetByteTools::Core::CException;
using JetByteTools::Core::_tstring;
using JetByteTools::Core::CStringConverter;
using JetByteTools::Core::ToString;
using JetByteTools::Core::IProvideUserData;
using JetByteTools::Core::IIndexedOpaqueUserData;
using JetByteTools::Core::IThreadPoolWorkerThreadCallback;
using JetByteTools::Core::GetStringLength;
using JetByteTools::Core::ILockableObject;

using JetByteTools::IO::IBuffer;
using JetByteTools::IO::CBufferChain;
using JetByteTools::IO::CSmartBuffer;
using JetByteTools::IO::CCustomPoolSmartBuffer;
using JetByteTools::IO::IAllocateBuffers;

using JetByteTools::Socket::IStreamSocket;
using JetByteTools::Socket::IAddress;
using JetByteTools::Socket::CAddressRenderer;
using JetByteTools::Socket::CSmartStreamSocket;
using JetByteTools::Socket::IAllocateSequencedStreamSockets;


using std::min;

///////////////////////////////////////////////////////////////////////////////
// CThreadPool
///////////////////////////////////////////////////////////////////////////////

CThreadPool::CThreadPool(
   IAllocateSequencedStreamSockets &socketAllocator,
   IAllocateBuffers &bufferAllocator,
   const ThreadCount initialThreads,
   const ThreadCount minThreads,
   const ThreadCount maxThreads,
   const ThreadCount maxDormantThreads,
   const Milliseconds poolMaintPeriod,
   const Milliseconds dispatchTimeout,
   const long numReads,
   const bool usingSequenceNumbers,
   const bool useMultipleWrites)
   :  m_threadPool(
         *this,
         initialThreads,
         minThreads,
         maxThreads,
         maxDormantThreads,
         poolMaintPeriod,
         dispatchTimeout,
         *this),
      m_userDataIndex(socketAllocator.RequestUserDataSlot(_T("CThreadPool::PerConnectionDataIndex"))),
      m_bufferChainIndex(bufferAllocator.RequestUserDataSlot(_T("CThreadPool::BufferChain"))),
      m_numReads(numReads),
      m_usingSequenceNumbers(usingSequenceNumbers),
      m_useMultipleWrites(useMultipleWrites),
      m_writeAsBytes(true),
      m_processHow(ProcessAlternate),
      m_processContiguouslyThisTime(true),
      m_useMultipleWritesThisTime(false),
      m_writeAsBytesThisTime(false)
{
}

void CThreadPool::Start()
{
   m_threadPool.Start();
}

void CThreadPool::WaitForShutdownToComplete(
   Milliseconds /*forceTerminationAfterMillis*/)
{
   m_threadPool.WaitForShutdownToComplete(/*forceTerminationAfterMillis*/);
}

void CThreadPool::DispatchConnectionEstablished(
   IStreamSocket &socket,
   const IAddress &address)
{
   // Allocate per connection data

   CPerConnectionData *pData = new CPerConnectionData(
      m_bufferChainIndex,
      CAddressRenderer::AsString(address, true));

   // And store our data in the socket...

   socket.SetUserPointer(m_userDataIndex, pData);

   DoDispatch(socket, 0, ConnectionEstablished);
}

void CThreadPool::DispatchCompleteMessage(
   IStreamSocket &socket,
   CSmartBuffer &buffer)
{
   DoDispatch(socket, buffer.Detach(), ReadCompleted);
}

void CThreadPool::DoDispatch(
   IStreamSocket &socket,
   IBuffer *pBuffer,
   const DispatchEvents event)
{
   socket.AddRef();

   if (!pBuffer && event != ConnectionEstablished)
   {
      throw CException(
         _T("CThreadPool::DoDispatch()"),
         _T("Unexpected: Only ConnectionEstablished events can have a null pBuffer"));
   }

   m_threadPool.Dispatch(
      reinterpret_cast<ULONG_PTR>(&socket),
      event,
      pBuffer);
}

void CThreadPool::OnSocketReleased(
   IIndexedOpaqueUserData &userData)
{
   // release per connection data

   CPerConnectionData &data = GetPerConnectionData(userData);

   userData.SetUserPointer(m_userDataIndex, 0);

   delete &data;
}

CThreadPool::CPerConnectionData &CThreadPool::GetPerConnectionData(
   const IIndexedOpaqueUserData &userData) const
{
   return *static_cast<CPerConnectionData*>(userData.GetUserPointer(m_userDataIndex));
}

void CThreadPool::OnConnectionEstablished(
   IStreamSocket &socket)
{
   std::string welcomeMessage("Welcome to large packet echo server\r\n");

   socket.Write(welcomeMessage.c_str(), GetStringLength<DWORD>(welcomeMessage));

   bool ok = true;

   for (long i = 0; ok && i < m_numReads; ++i)
   {
      ok = socket.TryRead();
   }
}

void CThreadPool::ProcessMessage(
   IStreamSocket &socket,
   CSmartBuffer &buffer_)
{
   CPerConnectionData &data = GetPerConnectionData(socket);

   data.UseConnection();

   CCustomPoolSmartBuffer buffer(m_usingSequenceNumbers ? data.m_outOfSequenceReads.GetNext(buffer_) : buffer_, socket);

   while (buffer.Get())
   {
      Output(_T("EchoMessage - ") + ToString(buffer->GetSequenceNumber()));

      // Reconstitute the buffer chain that makes up the complete message...

      CBufferChain buffers;

      buffers.Attach(buffer.Detach());

      if (m_processHow == ProcessContiguously ||
         (m_processHow == ProcessAlternate && m_processContiguouslyThisTime))
      {
         // if we need to process the data as one contiguous block of memory then we need to
         // do this...

         CSmartBuffer contiguousMessage(buffers.RemoveAsSingleBuffer());

         InternalProcessMessage(socket, contiguousMessage);

         m_processContiguouslyThisTime = false;
      }
      else
      {
         // if we can process it as discrete buffers of arbitrary size then we can do this

         InternalProcessMessage(socket, buffers);

         m_processContiguouslyThisTime = true;
      }

      if (m_usingSequenceNumbers)
      {
         buffer = data.m_outOfSequenceReads.ProcessAndGetNext();
      }
      else
      {
         buffer.Detach();
      }
   }
}

void CThreadPool::InternalProcessMessage(
   IStreamSocket &socket,
   CSmartBuffer &buffer)
{
   if (m_useMultipleWrites && m_useMultipleWritesThisTime)
   {
      CPerConnectionData &data = GetPerConnectionData(socket);

      ILockableObject::Owner lock(data.GetLock());

      IBuffer::BufferSize bytesSent = 0;

      const BYTE *pMessageData = buffer->GetMemory();
      const IBuffer::BufferSize messageDataLength = buffer->GetUsed();

      while (bytesSent < messageDataLength)
      {
         const IBuffer::BufferSize responsePacketSize = 666;

         const IBuffer::BufferSize bytesToSend = min(responsePacketSize, messageDataLength - bytesSent);

         socket.TryWrite(pMessageData + bytesSent, bytesToSend);

         bytesSent += bytesToSend;
      }

      m_useMultipleWritesThisTime = false;
   }
   else
   {
      if (m_writeAsBytes && m_writeAsBytesThisTime)
      {
         const BYTE *pData = buffer->GetMemory();

         const IBuffer::BufferSize dataLength = buffer->GetUsed();

         socket.TryWrite(pData, dataLength);

         m_writeAsBytesThisTime = false;
      }
      else
      {
         // Echo packet with a single write.

         socket.TryWrite(buffer);

         m_writeAsBytesThisTime = true;
      }

      m_useMultipleWritesThisTime = true;
   }
}

void CThreadPool::InternalProcessMessage(
   IStreamSocket &socket,
   CBufferChain &buffers)
{
   if (m_useMultipleWrites && m_useMultipleWritesThisTime)
   {
      CPerConnectionData &data = GetPerConnectionData(socket);

      ILockableObject::Owner lock(data.GetLock());

      CSmartBuffer buffer(buffers.GetNext());

      while (buffer.Get())
      {
         IBuffer::BufferSize bytesSent = 0;

         const BYTE *pMessageData = buffer->GetMemory();
         const IBuffer::BufferSize messageDataLength = buffer->GetUsed();

         while (bytesSent < messageDataLength)
         {
            const IBuffer::BufferSize responsePacketSize = 666;

            const IBuffer::BufferSize bytesToSend = min(responsePacketSize, messageDataLength - bytesSent);

            socket.TryWrite(pMessageData + bytesSent, bytesToSend);

            bytesSent += bytesToSend;
         }

         buffer = buffers.GetNext();
      }

      m_useMultipleWritesThisTime = false;
   }
   else
   {
      // Echo packet with a single write.
      socket.TryWrite(buffers);

      m_useMultipleWritesThisTime = true;
   }
}

IThreadPoolWorkerThreadCallback &CThreadPool::CreateWorkerThread()
{
   return *this;
}

void CThreadPool::DestroyWorkerThread(
   IThreadPoolWorkerThreadCallback & /*thread*/)
{

}

void CThreadPool::DisposeOfQueuedItemsAfterShutdown(
   const ULONG_PTR completionKey,
   const DWORD operation,
   OVERLAPPED *pOverlapped)
{
   Output(_T("DisposeOfQueuedItemsAfterShutdown"));

   reinterpret_cast<IStreamSocket *>(completionKey)->Release();

   if (operation == ReadCompleted)
   {
      // The destruction of the buffer chain will destroy the buffers
      // in the chain...

      CBufferChain buffers;

      buffers.Attach(static_cast<IBuffer *>(pOverlapped));
   }
}

bool CThreadPool::Initialise()
{
   Output(_T("Initialise"));

   return true;
}

void CThreadPool::Process(
   const ULONG_PTR completionKey,
   const DWORD operation,
   OVERLAPPED *pOverlapped)
{
   CSmartStreamSocket socket(reinterpret_cast<IStreamSocket *>(completionKey));

   try
   {
      switch(operation)
      {
         case ConnectionEstablished :

            OnConnectionEstablished(socket.GetRef());

         break;

         case ReadCompleted :
         {
            CCustomPoolSmartBuffer buffer(static_cast<IBuffer *>(pOverlapped), socket.GetRef());

            ProcessMessage(socket.GetRef(), buffer);
         }
         break;

         default:

            throw CException(
               _T("CThreadPool::Process()"),
               _T("Unexpected operation: ") + ToString(operation));
      }
   }
   catch(const CException &e)
   {
      OutputEx(_T("Process - Exception - ") + e.GetDetails());

      socket->AbortConnection();
   }
   catch(...)
   {
      OutputEx(_T("Process - Unexpected exception"));

      socket->AbortConnection();
   }
}

void CThreadPool::Shutdown()
{
   Output(_T("Shutdown"));
}

void CThreadPool::OnThreadPoolThreadStarted()
{
   Output(_T("OnThreadStarted"));
}

void CThreadPool::OnThreadPoolDispatch()
{
}

void CThreadPool::OnThreadPoolProcessingStarted(
   const bool /*dataAvailable*/)
{
   Output(_T("OnProcessingStarted"));
}

void CThreadPool::OnThreadPoolProcessingStopped(
   const bool /*dataAvailable*/)
{
   Output(_T("OnProcessingStopped"));
}

void CThreadPool::OnThreadPoolThreadStopped()
{
   Output(_T("OnThreadStopped"));
}

///////////////////////////////////////////////////////////////////////////////
// CThreadPool::CPerConnectionData
///////////////////////////////////////////////////////////////////////////////

CThreadPool::CPerConnectionData::CPerConnectionData(
   const IIndexedOpaqueUserData::UserDataIndex bufferChainIndex,
   const _tstring &address)
   :  m_bufferChain(bufferChainIndex),
      m_outOfSequenceReads(m_lock, m_bufferChain),
      m_numUses(0),
      m_address(address)
{
}

void CThreadPool::CPerConnectionData::UseConnection()
{
   ::InterlockedIncrement(&m_numUses);
}

_tstring CThreadPool::CPerConnectionData::GetConnectionDetails() const
{
   return m_address + _T(" - ") + ToString(m_numUses);
}

ILockableObject &CThreadPool::CPerConnectionData::GetLock()
{
   return m_lock;
}

///////////////////////////////////////////////////////////////////////////////
// End of file: ThreadPool.cpp
///////////////////////////////////////////////////////////////////////////////
