///////////////////////////////////////////////////////////////////////////////
// File: ThreadPool.cpp
/////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2004 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"
#include "JetByteTools/Admin/WinsockWrapper.h"

#include "ThreadPool.h"

#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/IProvideUserData.h"
#include "JetByteTools/CoreTools/IIndexedOpaqueUserData.h"
#include "JetByteTools/CoreTools/StringConverter.h"

#include "JetByteTools/IOTools/IBuffer.h"
#include "JetByteTools/IOTools/SmartBuffer.h"

#include "JetByteTools/SocketTools/IStreamSocket.h"
#include "JetByteTools/SocketTools/AddressRenderer.h"
#include "JetByteTools/SocketTools/SmartStreamSocket.h"

#pragma hdrstop

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Milliseconds;

using JetByteTools::Core::Output;
using JetByteTools::Core::OutputEx;
using JetByteTools::Core::CException;
using JetByteTools::Core::_tstring;
using JetByteTools::Core::CStringConverter;
using JetByteTools::Core::ToString;
using JetByteTools::Core::IProvideUserData;
using JetByteTools::Core::IIndexedOpaqueUserData;
using JetByteTools::Core::IThreadPoolWorkerThreadCallback;
using JetByteTools::Core::GetStringLength;

using JetByteTools::IO::IBuffer;
using JetByteTools::IO::CSmartBuffer;

using JetByteTools::Socket::IStreamSocket;
using JetByteTools::Socket::IAddress;
using JetByteTools::Socket::CAddressRenderer;
using JetByteTools::Socket::CSmartStreamSocket;

///////////////////////////////////////////////////////////////////////////////
// CThreadPool
///////////////////////////////////////////////////////////////////////////////

CThreadPool::CThreadPool(
   IProvideUserData &userDataProvider,
   const ThreadCount initialThreads,
   const ThreadCount minThreads,
   const ThreadCount maxThreads,
   const ThreadCount maxDormantThreads,
   const Milliseconds poolMaintPeriod,
   const Milliseconds dispatchTimeout)
   :  m_threadPool(
         *this,
         initialThreads,
         minThreads,
         maxThreads,
         maxDormantThreads,
         poolMaintPeriod,
         dispatchTimeout,
         *this),
      m_userDataIndex(userDataProvider.RequestUserDataSlot(_T("CThreadPool::PerConnectionDataIndex")))
{
}

void CThreadPool::Start()
{
   m_threadPool.Start();
}

void CThreadPool::WaitForShutdownToComplete(
   Milliseconds /*forceTerminationAfterMillis*/)
{
   m_threadPool.WaitForShutdownToComplete(/*forceTerminationAfterMillis*/);
}

void CThreadPool::DispatchConnectionEstablished(
   IStreamSocket &socket,
   const IAddress &address)
{
   // Allocate per connection data

   CPerConnectionData *pData = new CPerConnectionData(CAddressRenderer::AsString(address, true));

   // And store our data in the socket...

   socket.SetUserPointer(m_userDataIndex, pData);

   DoDispatch(socket, 0, ConnectionEstablished);
}

void CThreadPool::DispatchReadCompleted(
   IStreamSocket &socket,
   CSmartBuffer &buffer)
{
   DoDispatch(socket, buffer.Detach(), ReadCompleted);
}

void CThreadPool::DoDispatch(
   IStreamSocket &socket,
   IBuffer *pBuffer,
   const DispatchEvents event)
{
   socket.AddRef();

   if (!pBuffer && event != ConnectionEstablished)
   {
      throw CException(
         _T("CThreadPool::DoDispatch()"),
         _T("Unexpected: Only ConnectionEstablished events can have a null pBuffer"));
   }

   m_threadPool.Dispatch(
      reinterpret_cast<ULONG_PTR>(&socket),
      event,
      pBuffer);
}

void CThreadPool::OnSocketReleased(
   IIndexedOpaqueUserData &userData)
{
   // release per connection data

   CPerConnectionData &data = GetPerConnectionData(userData);

   userData.SetUserPointer(m_userDataIndex, 0);

   delete &data;
}

CThreadPool::CPerConnectionData &CThreadPool::GetPerConnectionData(
   const IIndexedOpaqueUserData &userData) const
{
   return *static_cast<CPerConnectionData*>(userData.GetUserPointer(m_userDataIndex));
}

void CThreadPool::OnConnectionEstablished(
   IStreamSocket &socket)
{
   CPerConnectionData &data = GetPerConnectionData(socket);

   std::string welcomeMessage("+OK POP3 server ready - ");

   welcomeMessage = welcomeMessage + CStringConverter::TtoA(data.GetConnectionDetails()) + "\r\n";

   socket.Write(welcomeMessage.c_str(), GetStringLength<DWORD>(welcomeMessage));

   socket.TryRead();
}

void CThreadPool::ProcessMessage(
   IStreamSocket &socket,
   const CSmartBuffer &buffer) const
{
   std::string command = reinterpret_cast<const char*>(buffer->GetMemory());

   for (IBuffer::BufferSize  i = 0; i < command.length(); ++i)
   {
      command[i] = (char)toupper(command[i]);
   }

   bool ok = false;

   if (command.length() >= 5)
   {
      // cant be a valid command, unless it's more than 5 chars

      if (command.substr(0, 3) == "TOP")
      {
         ok = true;
      }
      else
      {
         const std::string cmd = command.substr(0, 4);

         if (cmd == "APOP" ||
             cmd == "DELE" ||
             cmd == "LIST" ||
             cmd == "NOOP" ||
             cmd == "PASS" ||
             cmd == "QUIT" ||
             cmd == "RETR" ||
             cmd == "RSET" ||
             cmd == "STAT" ||
             cmd == "UIDL" ||
             cmd == "USER")
         {
            ok = true;
         }
      }
   }

   CPerConnectionData &data = GetPerConnectionData(socket);

   data.UseConnection();

   if (ok)
   {
      // We understand, but we aren't really a POP3 server...

      std::string response("-ERR sorry, we understand what you want, but, we're just a fake POP3 server... ");

      response = response + CStringConverter::TtoA(data.GetConnectionDetails()) + "\r\n";

      // We take a while to reply because we do lots of processing...

      //::Sleep(5000);

      socket.Write(response.c_str(), GetStringLength<DWORD>(response));
   }
   else
   {
      std::string response("-ERR sorry, we dont know what you mean, bye... ");

      response = response + CStringConverter::TtoA(data.GetConnectionDetails()) + "\r\n";

      socket.Write(response.c_str(), GetStringLength<DWORD>(response));
      socket.Shutdown();
   }
}

IThreadPoolWorkerThreadCallback &CThreadPool::CreateWorkerThread()
{
   return *this;
}

void CThreadPool::DestroyWorkerThread(
   IThreadPoolWorkerThreadCallback & /*thread*/)
{

}

void CThreadPool::DisposeOfQueuedItemsAfterShutdown(
   const ULONG_PTR completionKey,
   const DWORD operation,
   OVERLAPPED *pOverlapped)
{
   Output(_T("DisposeOfQueuedItemsAfterShutdown"));

   reinterpret_cast<IStreamSocket *>(completionKey)->Release();

   if (operation == ReadCompleted)
   {
      static_cast<IBuffer *>(pOverlapped)->Release();
   }
}

bool CThreadPool::Initialise()
{
   Output(_T("Initialise"));

   return true;
}

void CThreadPool::Process(
   const ULONG_PTR completionKey,
   const DWORD operation,
   OVERLAPPED *pOverlapped)
{
   CSmartStreamSocket socket(reinterpret_cast<IStreamSocket *>(completionKey));

   try
   {
      switch(operation)
      {
         case ConnectionEstablished :

            OnConnectionEstablished(socket.GetRef());

         break;

         case ReadCompleted :
         {
            CSmartBuffer buffer(static_cast<IBuffer *>(pOverlapped));

            ProcessMessage(socket.GetRef(), buffer);
         }
         break;

         default:

            throw CException(
               _T("CThreadPool::Process()"),
               _T("Unexpected operation: ") + ToString(operation));
      }
   }
   catch(const CException &e)
   {
      OutputEx(_T("Process - Exception - ") + e.GetDetails());

      socket->AbortConnection();
   }
   catch(...)
   {
      OutputEx(_T("Process - Unexpected exception"));

      socket->AbortConnection();
   }
}

void CThreadPool::Shutdown()
{
   Output(_T("Shutdown"));
}

void CThreadPool::OnThreadPoolThreadStarted()
{
   Output(_T("OnThreadStarted"));
}

void CThreadPool::OnThreadPoolDispatch()
{
}

void CThreadPool::OnThreadPoolProcessingStarted(
   const bool /*dataAvailable*/)
{
   Output(_T("OnProcessingStarted"));
}

void CThreadPool::OnThreadPoolProcessingStopped(
   const bool /*dataAvailable*/)
{
   Output(_T("OnProcessingStopped"));
}

void CThreadPool::OnThreadPoolThreadStopped()
{
   Output(_T("OnThreadStopped"));
}

///////////////////////////////////////////////////////////////////////////////
// CThreadPool::CPerConnectionData
///////////////////////////////////////////////////////////////////////////////

CThreadPool::CPerConnectionData::CPerConnectionData(
   const _tstring &address)
   :  m_numUses(0),
      m_address(address)
{
}

void CThreadPool::CPerConnectionData::UseConnection()
{
   ::InterlockedIncrement(&m_numUses);
}

_tstring CThreadPool::CPerConnectionData::GetConnectionDetails() const
{
   return m_address + _T(" - ") + ToString(m_numUses);
}

///////////////////////////////////////////////////////////////////////////////
// End of file: ThreadPool.cpp
///////////////////////////////////////////////////////////////////////////////
